@@ -1,10 +1,12 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE Rank2Types #-}
 {-# LANGUAGE EmptyDataDecls #-}
 {-# LANGUAGE ExistentialQuantification #-}
+{-# LANGUAGE GADTs #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE StandaloneDeriving #-}
 ---------------------------------------------------------
 -- |
 -- Copyright   : (c) 2006-2016, alpheccar.org
@@ -34,6 +36,7 @@ module Graphics.PDF.Draw(
 -- , writeCmd
  , runDrawing
  , setResource
+ , registerResource
  , emptyEnvironment
  , PDFXForm
  , PDFXObject(..)
@@ -72,6 +75,26 @@ module Graphics.PDF.Draw(
  , AnyAnnotation(..)
  , AnnotationStyle(..)
  , PDFShading(..)
+ , ColorSpace(..)
+ , colorSpaceName
+ , calculator1
+ , calculator2
+ , ColorFunction1(..)
+ , ColorFunction2(..)
+ , Function1(..)
+ , Function2(..)
+ , Global
+ , Local
+ , linearStitched
+ , FunctionObject(FunctionObject, FunctionStream)
+ , rsrcFromCalculator
+ , rsrcFromInterpolated
+ , rsrcFromSampled
+ , rsrcFromStitched
+ , ColorTuple
+ , domain1Dict
+ , domain2Dict
+ , SoftMask(..)
  , getRgbColor
  , emptyDrawState
  , Matrix(..)
@@ -81,8 +104,7 @@ module Graphics.PDF.Draw(
  , multiplyCurrentMatrixWith
  , PDFGlobals(..)
  ) where
- 
-import Data.Maybe
+
 #if !MIN_VERSION_base(4,8,0)
 import Data.Monoid
 #endif
@@ -91,6 +113,9 @@ import qualified Data.Map.Strict as M
 import qualified Data.IntMap as IM
 import qualified Data.Binary.Builder as BU
 import qualified Data.ByteString.Lazy as B
+import qualified Data.ByteString.Lazy.Char8 as C
+import qualified Data.Array as Array
+import Data.Array (Array)
 
 import Control.Monad.ST
 import Data.STRef
@@ -99,6 +124,8 @@ import Control.Monad.Writer.Class
 import Control.Monad.Reader.Class
 import Control.Monad.State
 
+import qualified Graphics.PDF.Expression as Expr
+import Graphics.PDF.Expression (PDFExpression)
 import Graphics.PDF.Coordinates
 import Graphics.PDF.LowLevel.Types
 import Graphics.PDF.LowLevel.Serializer
@@ -107,6 +134,8 @@ import Graphics.PDF.Data.PDFTree(PDFTree)
 import qualified Data.Text as T
 import Graphics.PDF.Fonts.Font(PDFFont(..))
 
+import Text.Printf (printf)
+
 data AnnotationStyle = AnnotationStyle !(Maybe Color)
 
 class AnnotationObject a where
@@ -149,6 +178,7 @@ data DrawState = DrawState {
                 ,  patterns :: M.Map (PDFReference AnyPdfPattern) String
                 ,  colorSpaces :: M.Map PDFColorSpace String
                 ,  shadings :: M.Map PDFShading String
+                ,  softMasks :: M.Map SoftMask String
                 ,  matrix :: [Matrix]
                 }
 data DrawEnvironment = DrawEnvironment {
@@ -222,8 +252,13 @@ modifyDrawST :: (forall s. DrawTuple s -> STRef s a) -> (a -> a) -> Draw ()
 modifyDrawST f g = Draw $ \env -> modifySTRef (f env) g
 
 -- | A PDF stream object
-data PDFStream = PDFStream !BU.Builder !Bool !(PDFReference MaybeLength) !PDFDictionary
-                                   
+data PDFStream =
+    PDFStream
+        !BU.Builder
+        !Bool
+        !(Either (PDFReference MaybeLength) PDFLength)
+        !PDFDictionary
+
 instance PdfObject PDFStream where
   toPDF (PDFStream s c l d) = 
       mconcat   $ [ toPDF dict
@@ -234,13 +269,17 @@ instance PdfObject PDFStream where
                   , serialize "endstream"]
    where
       compressedStream False = []
-      compressedStream True = if not (pdfDictMember (PDFName "Filter") d) then [(PDFName "Filter",AnyPdfObject $ [AnyPdfObject . PDFName $ "FlateDecode"])] else []
-      lenDict = PDFDictionary. M.fromList $ [ (PDFName "Length",AnyPdfObject l)] ++ compressedStream c
+      compressedStream True = if not (pdfDictMember (PDFName "Filter") d) then [entry "Filter" [PDFName $ "FlateDecode"]] else []
+      lenDict = dictFromList $ [either (entry "Length") (entry "Length") l] ++ compressedStream c
       dict = pdfDictUnion lenDict d
 
 instance PdfLengthInfo PDFStream where 
-  pdfLengthInfo (PDFStream s _ l _) = Just (B.length . BU.toLazyByteString $ s,l)
-    
+  pdfLengthInfo (PDFStream s _ el _) =
+      either
+        (\ref -> Just (PDFLength . B.length . BU.toLazyByteString $ s, ref))
+        (const Nothing)
+        el
+
 -- | An empty drawing
 emptyDrawing :: Draw ()
 emptyDrawing = return ()
@@ -259,7 +298,7 @@ supplyName = do
 emptyDrawState :: Int -> DrawState
 emptyDrawState ref = 
     let names = (map (("O" ++ (show ref)) ++ ) $ [replicate k ['a'..'z'] | k <- [1..]] >>= sequence) in
-    DrawState names emptyRsrc M.empty M.empty M.empty M.empty emptyDictionary []  M.empty M.empty M.empty [identity]
+    DrawState names emptyRsrc M.empty M.empty M.empty M.empty emptyDictionary []  M.empty M.empty M.empty M.empty [identity]
   
 -- | Execute the drawing commands to get a new state and an uncompressed PDF stream
 runDrawing :: Draw a -> DrawEnvironment -> DrawState -> (a,DrawState,BU.Builder)
@@ -318,6 +357,20 @@ setResource dict values oldCache = do
              return (newName,M.insert values newName oldCache)
         Just n -> return (n,oldCache)
 
+-- ToDo: setter and getter could be replaced by an Accessor or a Lens
+registerResource ::
+    (Ord a, PdfResourceObject a) =>
+    String ->
+    (DrawState -> M.Map a String) ->
+    (M.Map a String -> DrawState -> DrawState) ->
+    a -> Draw String
+registerResource dict getMap setMap resource = do
+    oldMap <- gets getMap
+    (newName,newMap) <- setResource dict resource oldMap
+    modifyStrict $ setMap newMap
+    return newName
+
+
 instance PDFGlobals Draw where
     bounds (PDFReference r) = getBoundInDraw r
     
@@ -330,9 +383,10 @@ class PDFXObject a where
     
     privateDrawXObject :: PDFReference a -> Draw ()
     privateDrawXObject (PDFReference r) = do
-        xobjectMap <- gets xobjects
-        (newName,newMap) <- setResource "XObject" (PDFReference r) xobjectMap
-        modifyStrict $ \s -> s { xobjects = newMap }
+        newName <-
+            registerResource "XObject"
+                xobjects (\newMap s -> s { xobjects = newMap })
+                (PDFReference r)
         tell . mconcat  $ [ serialize "\n/" 
                           , serialize newName
                           , serialize " Do"
@@ -520,10 +574,10 @@ data PDFViewerPreferences = PDFViewerPreferences { hideToolbar :: Bool -- ^ To h
 data PDFOutline = PDFOutline !(PDFReference PDFOutlineEntry) !(PDFReference PDFOutlineEntry)
 
 instance PdfObject PDFOutline where
- toPDF (PDFOutline first lasto) = toPDF $ PDFDictionary. M.fromList $ [
-    (PDFName "Type",AnyPdfObject . PDFName $ "Outlines")
-  , (PDFName "First",AnyPdfObject first)
-  , (PDFName "Last",AnyPdfObject lasto)
+ toPDF (PDFOutline first lasto) = toPDF $ dictFromList $ [
+    entry "Type" (PDFName $ "Outlines")
+  , entry "First" first
+  , entry "Last" lasto
   ]
 
 instance PdfLengthInfo PDFOutline where
@@ -562,14 +616,14 @@ data OutlineCtx a = Top | Child { value :: a
 data OutlineLoc  a = OutlineLoc (Tree a) (OutlineCtx a)
 
 instance PdfObject PDFViewerPreferences where
-  toPDF (PDFViewerPreferences ht hm hwui fw cw ddt nfspm ) = toPDF $ PDFDictionary. M.fromList $ 
-   [ (PDFName "HideToolbar",AnyPdfObject ht)
-   , (PDFName "HideMenubar",AnyPdfObject hm)
-   , (PDFName "HideWindowUI",AnyPdfObject hwui)
-   , (PDFName "FitWindow",AnyPdfObject fw)
-   , (PDFName "CenterWindow",AnyPdfObject cw)
-   , (PDFName "DisplayDocTitle",AnyPdfObject ddt)
-   , (PDFName "NonFullScreenPageMode",AnyPdfObject  . PDFName . show $ nfspm)
+  toPDF (PDFViewerPreferences ht hm hwui fw cw ddt nfspm ) = toPDF $ dictFromList $
+   [ entry "HideToolbar" ht
+   , entry "HideMenubar" hm
+   , entry "HideWindowUI" hwui
+   , entry "FitWindow" fw
+   , entry "CenterWindow" cw
+   , entry "DisplayDocTitle" ddt
+   , entry "NonFullScreenPageMode" (PDFName . show $ nfspm)
    ]
 
 instance PdfLengthInfo PDFViewerPreferences where
@@ -584,20 +638,20 @@ instance Show PDFTransStyle where
    show (Glitter _) = "Glitter"
 
 instance PdfObject PDFTransition where
- toPDF (PDFTransition d t) = toPDF $ PDFDictionary. M.fromList $ 
-   [ (PDFName "Type",AnyPdfObject (PDFName "Trans"))
-   , (PDFName "S",AnyPdfObject (PDFName (show t)))
-   , (PDFName "D",AnyPdfObject d)
+ toPDF (PDFTransition d t) = toPDF $ dictFromList $
+   [ entry "Type" (PDFName "Trans")
+   , entry "S" (PDFName (show t))
+   , entry "D" d
    ] ++ optionalDm t ++ optionalM t ++ optionalDi t
   where
-    optionalDm (Split a _) = [ (PDFName "Dm",AnyPdfObject (PDFName (show a)))]
-    optionalDm (Blinds a) = [ (PDFName "Dm",AnyPdfObject (PDFName (show a)))]
+    optionalDm (Split a _) = [ entry "Dm" (PDFName (show a))]
+    optionalDm (Blinds a) = [ entry "Dm" (PDFName (show a))]
     optionalDm _ = []
-    optionalM (Split _ a) = [ (PDFName "M",AnyPdfObject (PDFName (show a)))]
-    optionalM (Box a) = [ (PDFName "M",AnyPdfObject (PDFName (show a)))]
+    optionalM (Split _ a) = [ entry "M" (PDFName (show a))]
+    optionalM (Box a) = [ entry "M" (PDFName (show a))]
     optionalM _ = []    
-    optionalDi (Wipe a) = [ (PDFName "Di",AnyPdfObject (floatDirection a))]
-    optionalDi (Glitter a)  = [ (PDFName "Di",AnyPdfObject (floatDirection a))]
+    optionalDi (Wipe a) = [ entry "Di" (floatDirection a)]
+    optionalDi (Glitter a)  = [ entry "Di" (floatDirection a)]
     optionalDi _ = []  
 
 instance PdfLengthInfo PDFTransition where
@@ -605,35 +659,33 @@ instance PdfLengthInfo PDFTransition where
 -- PDF Pages
 
 instance PdfObject PDFPages where
- toPDF (PDFPages c Nothing l) = toPDF $ PDFDictionary. M.fromList $ 
-  [ (PDFName "Type",AnyPdfObject (PDFName "Pages"))
-  , (PDFName "Kids",AnyPdfObject $ map AnyPdfObject l)
-  , (PDFName "Count",AnyPdfObject . PDFInteger $ c)
-  ] 
- toPDF (PDFPages c (Just theParent) l) = toPDF $ PDFDictionary. M.fromList $ 
-  [ (PDFName "Type",AnyPdfObject (PDFName "Pages"))
-  , (PDFName "Parent",AnyPdfObject theParent)
-  , (PDFName "Kids",AnyPdfObject $ map AnyPdfObject l)
-  , (PDFName "Count",AnyPdfObject . PDFInteger $ c)
-  ] 
+ toPDF (PDFPages c Nothing l) = toPDF $ dictFromList $
+  [ entry "Type" (PDFName "Pages")
+  , entry "Kids" l
+  , entry "Count" (PDFInteger $ c)
+  ]
+ toPDF (PDFPages c (Just theParent) l) = toPDF $ dictFromList $
+  [ entry "Type" (PDFName "Pages")
+  , entry "Parent" theParent
+  , entry "Kids" l
+  , entry "Count" (PDFInteger $ c)
+  ]
 
 instance PdfLengthInfo PDFPages where
 
 
 instance PdfObject PDFPage where
- toPDF (PDFPage (Just theParent) box content theRsrc d t theAnnots) = toPDF $ PDFDictionary. M.fromList $ 
-  [ (PDFName "Type",AnyPdfObject (PDFName "Page"))
-  , (PDFName "Parent",AnyPdfObject theParent)
-  , (PDFName "MediaBox",AnyPdfObject box)
-  , (PDFName "Contents",AnyPdfObject content)
-  , if isJust theRsrc 
-      then
-       (PDFName "Resources",AnyPdfObject . fromJust $ theRsrc) 
-      else 
-       (PDFName "Resources",AnyPdfObject emptyDictionary)
-  ] ++ (maybe [] (\x -> [(PDFName "Dur",AnyPdfObject x)]) d)
-  ++ (maybe [] (\x -> [(PDFName "Trans",AnyPdfObject x)]) t)
-  ++ ((\x -> if null x then [] else [(PDFName "Annots",AnyPdfObject x)]) theAnnots)
+ toPDF (PDFPage (Just theParent) box content theRsrc d t theAnnots) = toPDF $ dictFromList $
+  [ entry "Type" (PDFName "Page")
+  , entry "Parent" theParent
+  , entry "MediaBox" box
+  , entry "Contents" content
+  , case theRsrc of
+      Just res -> entry "Resources" res
+      Nothing -> entry "Resources" emptyDictionary
+  ] ++ (maybe [] (\x -> [entry "Dur" x]) d)
+  ++ (maybe [] (\x -> [entry "Trans" x]) t)
+  ++ ((\x -> if null x then [] else [entry "Annots" x]) theAnnots)
  toPDF (PDFPage Nothing _ _ _ _ _ _) = noPdfObject
 
 instance PdfLengthInfo PDFPage where
@@ -641,13 +693,13 @@ instance PdfLengthInfo PDFPage where
 -- Main objects in a PDF document
 
 instance PdfObject PDFCatalog where
- toPDF (PDFCatalog outlines lPages pgMode pgLayout viewerPrefs) = toPDF $ PDFDictionary . M.fromList $ 
-   [ (PDFName "Type",AnyPdfObject (PDFName "Catalog"))
-   , (PDFName "Pages",AnyPdfObject lPages)
-   , (PDFName "PageMode", AnyPdfObject . PDFName . show $ pgMode)
-   , (PDFName "PageLayout", AnyPdfObject . PDFName . show $ pgLayout)
-   , (PDFName "ViewerPreferences", AnyPdfObject viewerPrefs)
-   ] ++ (maybe [] (\x -> [(PDFName "Outlines",AnyPdfObject x)]) outlines)
+ toPDF (PDFCatalog outlines lPages pgMode pgLayout viewerPrefs) = toPDF $ dictFromList $
+   [ entry "Type" (PDFName "Catalog")
+   , entry "Pages" lPages
+   , entry "PageMode" (PDFName . show $ pgMode)
+   , entry "PageLayout" (PDFName . show $ pgLayout)
+   , entry "ViewerPreferences" viewerPrefs
+   ] ++ (maybe [] (\x -> [entry "Outlines" x]) outlines)
 
 instance PdfLengthInfo PDFCatalog where
 
@@ -660,23 +712,23 @@ instance PdfLengthInfo OutlineStyle where
 
 instance PdfObject PDFOutlineEntry where
  toPDF (PDFOutlineEntry title theParent prev next first theLast count dest color style) = 
-     toPDF $ PDFDictionary. M.fromList $ [
-        (PDFName "Title",AnyPdfObject title)
-        , (PDFName "Parent",AnyPdfObject theParent)
+     toPDF $ dictFromList $ [
+        entry "Title" title
+        , entry "Parent" theParent
         ]
       ++
-      maybe [] (\x -> [(PDFName "Prev",AnyPdfObject x)]) prev
+      maybe [] (\x -> [entry "Prev" x]) prev
       ++
-      maybe [] (\x -> [(PDFName "Next",AnyPdfObject x)]) next
+      maybe [] (\x -> [entry "Next" x]) next
       ++
-      maybe [] (\x -> [(PDFName "First",AnyPdfObject x)]) first
+      maybe [] (\x -> [entry "First" x]) first
       ++
-      maybe [] (\x -> [(PDFName "Last",AnyPdfObject x)]) theLast
+      maybe [] (\x -> [entry "Last" x]) theLast
       ++
-      [ (PDFName "Count",AnyPdfObject (PDFInteger count))
-      , (PDFName "Dest",AnyPdfObject dest)
-      , (PDFName "C",AnyPdfObject color)
-      , (PDFName "F",AnyPdfObject style)
+      [ entry "Count" (PDFInteger count)
+      , entry "Dest" dest
+      , entry "C" color
+      , entry "F" style
       ]
 
 instance PdfLengthInfo PDFOutlineEntry where
@@ -691,9 +743,9 @@ instance PdfLengthInfo Destination where
 
                                               
 instance PdfObject Color where
-   toPDF (Rgb r g b) = toPDF . map AnyPdfObject $ [r,g,b]  
+   toPDF (Rgb r g b) = toPDF [r,g,b]
    toPDF (Hsv h s v) = let (r,g,b) = hsvToRgb (h,s,v)
-    in toPDF . map AnyPdfObject $ [r,g,b]
+    in toPDF [r,g,b]
 
 instance PdfLengthInfo Color where
 
@@ -726,57 +778,363 @@ getRgbColor :: Color -> (PDFFloat,PDFFloat,PDFFloat)
 getRgbColor (Rgb r g b) = (r, g, b)  
 getRgbColor (Hsv h s v) = let (r,g,b) = hsvToRgb (h,s,v) in (r, g, b)  
 
+type FloatRGB  = (PDFFloat, PDFFloat, PDFFloat)
+type FloatCMYK = (PDFFloat, PDFFloat, PDFFloat, PDFFloat)
+
+class ColorTuple a where
+    rgbHex :: a -> String
+    colorComponents :: a -> [PDFFloat]
+    colorDimensions :: f a e -> Int
+
+instance ColorTuple PDFFloat where
+    rgbHex c = printf "%02X" (byteFromFloat c)
+    colorComponents c = [c]
+    colorDimensions _ = 1
+
+instance (a ~ PDFFloat, b ~ PDFFloat, c ~ PDFFloat) => ColorTuple (a,b,c) where
+    rgbHex (r,g,b) =
+        printf "%02X%02X%02X"
+            (byteFromFloat r) (byteFromFloat g) (byteFromFloat b)
+    colorComponents (r,g,b) = [r,g,b]
+    colorDimensions _ = 3
+
+instance
+    (a ~ PDFFloat, b ~ PDFFloat, c ~ PDFFloat, d ~ PDFFloat) =>
+        ColorTuple (a,b,c,d) where
+    rgbHex (c,m,y,k) =
+        printf "%02X%02X%02X%02X"
+            (byteFromFloat c) (byteFromFloat m)
+            (byteFromFloat y) (byteFromFloat k)
+    colorComponents (c,m,y,k) = [c,m,y,k]
+    colorDimensions _ = 4
+
+byteFromFloat :: PDFFloat -> Int
+byteFromFloat x = round $ min 255 $ max 0 $ x*255
+
+pdfStreamFromLazyByteString :: C.ByteString -> PDFDictionary -> PDFStream
+pdfStreamFromLazyByteString stream dict =
+    PDFStream
+        (BU.fromLazyByteString stream)
+        False
+        (Right . PDFLength . B.length $ stream)
+        dict
+
+data FunctionObject a e =
+      FunctionObject (PDFReference PDFDictionary)
+    | FunctionStream (PDFReference PDFStream)
+    deriving (Eq, Ord)
+
+rsrcFromSampled ::
+    (ColorTuple a) =>
+    PDFDictionary ->
+    ((i, i) -> [Int]) -> Array i a -> PDFStream
+rsrcFromSampled domain computeSizes arr =
+    pdfStreamFromLazyByteString
+        ((C.pack $ concatMap rgbHex $ Array.elems arr)
+            <>
+            C.pack " >")
+        (pdfDictUnion domain . dictFromList $
+           [entry "FunctionType" (PDFInteger 0),
+            entry "Size" (computeSizes $ Array.bounds arr),
+            entry "BitsPerSample" (PDFInteger 8),
+            entry "Filter" (PDFName "ASCIIHexDecode")
+            ])
+
 -- | Interpolation function
-interpole :: Int -> PDFFloat -> PDFFloat -> AnyPdfObject
-interpole n x y = AnyPdfObject . PDFDictionary . M.fromList $ 
-                            [ (PDFName "FunctionType", AnyPdfObject . PDFInteger $ 2)
-                            , (PDFName "Domain", AnyPdfObject . map AnyPdfObject $ ([0,1] :: [PDFFloat]))
-                            , (PDFName "C0", AnyPdfObject . map AnyPdfObject $ [x])
-                            , (PDFName "C1", AnyPdfObject . map AnyPdfObject $ [y])
-                            , (PDFName "N", AnyPdfObject . PDFInteger $  n)
+rsrcFromInterpolated ::
+    (ColorTuple a) =>
+    PDFDictionary -> PDFFloat -> a -> a -> PDFDictionary
+rsrcFromInterpolated domain n a b =
+    pdfDictUnion domain . dictFromList $
+                            [ entry "FunctionType" (PDFInteger 2)
+                            , entry "C0" (colorComponents a)
+                            , entry "C1" (colorComponents b)
+                            , entry "N" n
                             ]
 
--- | A shading                             
-data PDFShading = AxialShading PDFFloat PDFFloat PDFFloat PDFFloat Color Color
-                | RadialShading PDFFloat PDFFloat PDFFloat PDFFloat PDFFloat PDFFloat Color Color
+rsrcFromStitched ::
+    (ColorTuple a, Expr.Result e) =>
+    PDFDictionary ->
+    Function1 Local a e -> [(PDFFloat, Function1 Local a e)] -> PDFDictionary
+rsrcFromStitched domain part parts =
+    let funcs = part : map snd parts in
+    pdfDictUnion domain . dictFromList $
+        [ entry "FunctionType" (PDFInteger 3)
+        , entry "Bounds" (map fst parts)
+        , entry "Encode" (concatMap (const [0,1::Int]) funcs)
+        , entry "Functions" (map toRsrc funcs)
+        ]
+
+rsrcFromCalculator ::
+    (Expr.Function f) => PDFDictionary -> f -> PDFStream
+rsrcFromCalculator domain f =
+    pdfStreamFromLazyByteString
+        (C.cons '{' $ C.snoc (Expr.serialize f) '}')
+        (pdfDictUnion domain . dictFromList $
+            [entry "FunctionType" (PDFInteger 4)])
+
+
+type ExprFloat = PDFExpression PDFFloat
+type ExprRGB = (ExprFloat, ExprFloat, ExprFloat)
+type ExprCMYK = (ExprFloat, ExprFloat, ExprFloat, ExprFloat)
+
+
+data ColorSpace a e where
+    GraySpace :: ColorSpace PDFFloat ExprFloat
+    RGBSpace :: ColorSpace FloatRGB ExprRGB
+    CMYKSpace :: ColorSpace FloatCMYK ExprCMYK
+
+deriving instance Eq (ColorSpace a e)
+deriving instance Ord (ColorSpace a e)
+
+colorSpaceName :: ColorSpace a e -> PDFName
+colorSpaceName space =
+    case space of
+        GraySpace -> PDFName "DeviceGray"
+        RGBSpace -> PDFName "DeviceRGB"
+        CMYKSpace -> PDFName "DeviceCMYK"
+
+colorSpaceEntry :: ColorSpace a e -> (PDFName, AnyPdfObject)
+colorSpaceEntry space = entry "ColorSpace" $ colorSpaceName space
+
+rangeEntry :: (ColorTuple a) => f a e -> (PDFName, AnyPdfObject)
+rangeEntry func =
+    entry "Range" $ concat $ replicate (colorDimensions func) [0,1::Int]
+
+
+data ColorFunction1 =
+    forall a e.
+    (ColorTuple a, Expr.Result e) =>
+    ColorFunction1 (ColorSpace a e) (Function1 Local a e)
+
+instance Eq ColorFunction1 where
+    ColorFunction1 spaceA funcA == ColorFunction1 spaceB funcB  =
+        case (spaceA, spaceB) of
+            (GraySpace, GraySpace) -> funcA == funcB
+            (RGBSpace, RGBSpace) -> funcA == funcB
+            (CMYKSpace, CMYKSpace) -> funcA == funcB
+            _ -> False
+
+instance Ord ColorFunction1 where
+    compare (ColorFunction1 spaceA funcA) (ColorFunction1 spaceB funcB) =
+        case (spaceA, spaceB) of
+            (GraySpace, GraySpace) -> compare funcA funcB
+            (RGBSpace, RGBSpace) -> compare funcA funcB
+            (CMYKSpace, CMYKSpace) -> compare funcA funcB
+            (GraySpace, _) -> LT; (_, GraySpace) -> GT
+            (RGBSpace,  _) -> LT; (_, RGBSpace)  -> GT
+
+
+data ColorFunction2 =
+    forall a e.
+    (ColorTuple a, Expr.Result e) =>
+    ColorFunction2 (ColorSpace a e) (Function2 Local a e)
+
+instance Eq ColorFunction2 where
+    ColorFunction2 spaceA funcA == ColorFunction2 spaceB funcB  =
+        case (spaceA, spaceB) of
+            (GraySpace, GraySpace) -> funcA == funcB
+            (RGBSpace, RGBSpace) -> funcA == funcB
+            (CMYKSpace, CMYKSpace) -> funcA == funcB
+            _ -> False
+
+instance Ord ColorFunction2 where
+    compare (ColorFunction2 spaceA funcA) (ColorFunction2 spaceB funcB) =
+        case (spaceA, spaceB) of
+            (GraySpace, GraySpace) -> compare funcA funcB
+            (RGBSpace, RGBSpace) -> compare funcA funcB
+            (CMYKSpace, CMYKSpace) -> compare funcA funcB
+            (GraySpace, _) -> LT; (_, GraySpace) -> GT
+            (RGBSpace,  _) -> LT; (_, RGBSpace)  -> GT
+
+
+data Global
+data Local
+
+data Function1 scope a e where
+    GlobalFunction1 ::
+        FunctionObject (PDFFloat -> a) (ExprFloat -> e) ->
+        Function1 Local a e
+    Sampled1 :: Array Int a -> Function1 Global a e
+    Interpolated1 :: PDFFloat -> a -> a -> Function1 scope a e
+    Stitched1 ::
+        Function1 Local a e -> [(PDFFloat, Function1 Local a e)] ->
+        Function1 scope a e
+    Calculator1 :: (ExprFloat -> e) -> Function1 Global a e
+
+instance
+    (Local ~ scope, ColorTuple a, Eq a, Expr.Result e) =>
+        Eq (Function1 scope a e) where
+    a==b =
+        case (a,b) of
+            (GlobalFunction1 fa, GlobalFunction1 fb) -> fa == fb
+            (Interpolated1 na xa ya, Interpolated1 nb xb yb) ->
+                (na, xa, ya) == (nb, xb, yb)
+            (Stitched1 partA partsA, Stitched1 partB partsB) ->
+                (partA, partsA) == (partB, partsB)
+            _ -> False
+
+instance
+    (Local ~ scope, ColorTuple a, Ord a, Expr.Result e) =>
+        Ord (Function1 scope a e) where
+    compare a b =
+        case (a,b) of
+            (GlobalFunction1 fa, GlobalFunction1 fb) -> compare fa fb
+            (Interpolated1 na xa ya, Interpolated1 nb xb yb) ->
+                compare (na, xa, ya) (nb, xb, yb)
+            (Stitched1 partA partsA, Stitched1 partB partsB) ->
+                compare (partA, partsA) (partB, partsB)
+            (GlobalFunction1 _, _) -> LT
+            (_, GlobalFunction1 _) -> GT
+            (Interpolated1 _ _ _, _) -> LT
+            (_, Interpolated1 _ _ _) -> GT
+
+
+-- ToDo: would be more type-safe with non-empty: NonEmpty (NonEmpty [])
+mapAdjacent :: (a -> a -> b) -> [a] -> [b]
+mapAdjacent f xs = zipWith f xs (tail xs)
+
+linearStitched ::
+    (ColorTuple a) => a -> [(PDFFloat, a)] -> a -> Function1 Local a e
+linearStitched firstY nodes lastY =
+    case mapAdjacent (Interpolated1 1)
+            (firstY : map snd nodes ++ lastY : []) of
+        [] -> error "list should be non-empty by construction"
+        part:parts -> Stitched1 part (zip (map fst nodes) parts)
+
+calculator1 :: (ExprFloat -> e) -> Function1 Global a e
+calculator1 = Calculator1
+
+domain1Dict :: (ColorTuple a) => f a e -> PDFDictionary
+domain1Dict func =
+    dictFromList [
+        entry "Domain" [0,1::Int],
+        rangeEntry func
+    ]
+
+instance
+    (Local ~ scope, ColorTuple a, Expr.Result e) =>
+        PdfResourceObject (Function1 scope a e) where
+    toRsrc func =
+        let domain = domain1Dict func in
+        case func of
+            GlobalFunction1 (FunctionObject obj) -> AnyPdfObject obj
+            GlobalFunction1 (FunctionStream obj) -> AnyPdfObject obj
+            Interpolated1 n x y ->
+                AnyPdfObject $ rsrcFromInterpolated domain n x y
+            Stitched1 part parts ->
+                AnyPdfObject $ rsrcFromStitched domain part parts
+
+
+data Function2 scope a e where
+    GlobalFunction2 ::
+        FunctionObject
+            (PDFFloat -> PDFFloat -> a) (ExprFloat -> ExprFloat -> e) ->
+        Function2 Local a e
+    Sampled2 :: (Array (Int,Int) a) -> Function2 Global a e
+    Calculator2 :: (ExprFloat -> ExprFloat -> e) -> Function2 Global a e
+
+instance
+    (Local ~ scope, ColorTuple a, Eq a, Expr.Result e) =>
+        Eq (Function2 scope a e) where
+    a==b =
+        case (a,b) of
+            (GlobalFunction2 fa, GlobalFunction2 fb) -> fa == fb
+
+instance
+    (Local ~ scope, ColorTuple a, Ord a, Expr.Result e) =>
+        Ord (Function2 scope a e) where
+    compare a b =
+        case (a,b) of
+            (GlobalFunction2 fa, GlobalFunction2 fb) -> compare fa fb
+
+calculator2 :: (ExprFloat -> ExprFloat -> e) -> Function2 Global a e
+calculator2 = Calculator2
+
+domain2Dict :: (ColorTuple a) => f a e -> PDFDictionary
+domain2Dict func =
+    dictFromList [
+        entry "Domain" [0,1, 0,1::Int],
+        rangeEntry func
+    ]
+
+instance
+    (Local ~ scope, ColorTuple a, Expr.Result e) =>
+        PdfResourceObject (Function2 scope a e) where
+    toRsrc func =
+        case func of
+            GlobalFunction2 (FunctionObject obj) -> AnyPdfObject obj
+            GlobalFunction2 (FunctionStream obj) -> AnyPdfObject obj
+
+
+-- | A shading
+data PDFShading =
+      FunctionalShading Matrix ColorFunction2
+    | AxialShading PDFFloat PDFFloat PDFFloat PDFFloat ColorFunction1
+    | RadialShading PDFFloat PDFFloat PDFFloat PDFFloat PDFFloat PDFFloat ColorFunction1
                 deriving(Eq,Ord)
 
+matrixCoefficients :: Matrix -> [PDFFloat]
+matrixCoefficients (Matrix a b c d e f) = [a,b,c,d,e,f]
+
 instance PdfResourceObject PDFShading where
-      toRsrc (AxialShading x0 y0 x1 y1 ca cb) = AnyPdfObject . PDFDictionary . M.fromList $
-                                 [ (PDFName "ShadingType",AnyPdfObject . PDFInteger $ 2)
-                                 , (PDFName "Coords",AnyPdfObject . map AnyPdfObject $ [x0,y0,x1,y1])
-                                 , (PDFName "ColorSpace",AnyPdfObject . PDFName $ "DeviceRGB")
-                                 , (PDFName "Function",AnyPdfObject $ [interpole 1 ra rb,interpole 1 ga gb,interpole 1 ba bb])
+      toRsrc (FunctionalShading mat (ColorFunction2 cs func)) =
+          AnyPdfObject . dictFromList $
+                                 [ entry "ShadingType" (PDFInteger $ 1)
+                                 , entry "Matrix" (matrixCoefficients $ mat)
+                                 , colorSpaceEntry cs
+                                 , entry "Function" (toRsrc func)
                                  ]
-        where
-            (ra,ga,ba) = getRgbColor ca
-            (rb,gb,bb) = getRgbColor cb
-      toRsrc (RadialShading x0 y0 r0 x1 y1 r1 ca cb) = AnyPdfObject . PDFDictionary . M.fromList $
-                                         [ (PDFName "ShadingType",AnyPdfObject . PDFInteger $ 3)
-                                         , (PDFName "Coords",AnyPdfObject . map AnyPdfObject $ [x0,y0,r0,x1,y1,r1])
-                                         , (PDFName "ColorSpace",AnyPdfObject . PDFName $ "DeviceRGB")
-                                         , (PDFName "Function",AnyPdfObject $ [interpole 1 ra rb,interpole 1 ga gb,interpole 1 ba bb])
-                                         ]
-        where
-           (ra,ga,ba) = getRgbColor ca
-           (rb,gb,bb) = getRgbColor cb
+      toRsrc (AxialShading x0 y0 x1 y1 (ColorFunction1 cs func)) =
+          AnyPdfObject . dictFromList $
+                                 [ entry "ShadingType" (PDFInteger $ 2)
+                                 , entry "Coords" [x0,y0,x1,y1]
+                                 , entry "Extend" [True, True]
+                                 , colorSpaceEntry cs
+                                 , entry "Function" (toRsrc func)
+                                 ]
+      toRsrc (RadialShading x0 y0 r0 x1 y1 r1 (ColorFunction1 cs func)) =
+          AnyPdfObject . dictFromList $
+                                 [ entry "ShadingType" (PDFInteger $ 3)
+                                 , entry "Coords" [x0,y0,r0,x1,y1,r1]
+                                 , entry "Extend" [True, True]
+                                 , colorSpaceEntry cs
+                                 , entry "Function" (toRsrc func)
+                                 ]
+
+
+newtype SoftMask = SoftMask (PDFReference PDFXForm)
+    deriving (Eq, Ord)
+
+instance PdfResourceObject SoftMask where
+    toRsrc (SoftMask ref) =
+        AnyPdfObject $
+        dictFromList $
+            entry "Type" (PDFName "ExtGState") :
+            entry "SMask"
+                (dictFromList $
+                    entry "Type" (PDFName "Mask") :
+                    entry "S" (PDFName "Luminosity") :
+                    entry "G" ref :
+                    []) :
+            []
 
 
 -- | Apply a transformation matrix to the current coordinate frame
 applyMatrix :: Matrix -> Draw ()
 applyMatrix m@(Matrix a b c d e f)  = do
     multiplyCurrentMatrixWith m
-    tell . mconcat $[ serialize '\n'
-                    , toPDF a
-                    , serialize ' '
-                    , toPDF b
-                    , serialize ' '
-                    , toPDF c
-                    , serialize ' '
-                    , toPDF d
-                    , serialize ' '
-                    , toPDF e
-                    , serialize ' '
-                    , toPDF f
-                    , serialize " cm"
-                    ]
+    tell . mconcat $ [ serialize '\n'
+                     , toPDF a
+                     , serialize ' '
+                     , toPDF b
+                     , serialize ' '
+                     , toPDF c
+                     , serialize ' '
+                     , toPDF d
+                     , serialize ' '
+                     , toPDF e
+                     , serialize ' '
+                     , toPDF f
+                     , serialize " cm"
+                     ]
