@@ -32,11 +32,11 @@ module Graphics.PDF.Image(
  ) where
      
 import Graphics.PDF.LowLevel.Types
-import qualified Data.Map.Strict as M
 import Graphics.PDF.Draw
 import Graphics.PDF.Resources
 import Graphics.PDF.Pages
 import qualified Data.ByteString.Lazy as B
+import Control.Monad
 import Control.Monad.Writer
 #if __GLASGOW_HASKELL__ >= 608
 import System.IO hiding(withFile)
@@ -78,8 +78,8 @@ m_sof0 :: Int
 m_sof0 = 0xc0 
 m_sof1 :: Int 
 m_sof1 = 0xc1 
---m_sof2 :: Int 
---m_sof2 = 0xc2  
+m_sof2 :: Int 
+m_sof2 = 0xc2  
 m_sof3 :: Int 
 m_sof3 = 0xc3  
 m_sof5 :: Int 
@@ -126,8 +126,8 @@ m_soi :: Int
 m_soi = 0xd8 
 m_eoi :: Int   
 m_eoi = 0xd9 
---m_sos :: Int   
---m_sos = 0xda
+m_sos :: Int   
+m_sos = 0xda
 --m_dqt :: Int    
 --m_dqt = 0xdb 
 --m_dnl :: Int   
@@ -247,14 +247,24 @@ parseJpegContent h = do
     case sof of
         a | a `elem` [m_sof5,m_sof6,m_sof7,m_sof9,m_sof10,m_sof11,m_sof13,m_sof14,m_sof15] ->
                 EXC.throwError "Unuspported compression mode"
-          | a `elem` [m_sof0,m_sof1,m_sof3] -> do
+          | a `elem` [m_sof0,m_sof1,m_sof2,m_sof3] -> do
               _ <- readWord16 h
               bits_per_component <- readWord8 h
               height <- readWord16 h
               width <- readWord16 h
               color_space  <- readWord8 h
               return (bits_per_component,height,width,color_space)                  
-          | a `elem` [m_soi,m_eoi,m_tem,m_rst0,m_rst1,m_rst2,m_rst3,m_rst4,m_rst5,m_rst6,m_rst7] -> parseJpegContent h
+          | a `elem` [m_soi,m_tem,m_rst0,m_rst1,m_rst2,m_rst3,m_rst4,m_rst5,m_rst6,m_rst7] -> parseJpegContent h
+          | a == m_sos -> let
+            loop = do
+              x <- readWord8 h
+              if x /= 0xff then loop else do
+                y <- readWord8 h
+                if y == 0x00 then loop else do
+                  io $ hSeek h RelativeSeek (-2)
+                  parseJpegContent h
+            in loop
+          | a == m_eoi -> EXC.throwError "parseJpegContent: hit end of image (EOI) marker before getting JPEG metadata"
           | otherwise -> do
                l <- readWord16 h
                io $ hSeek h RelativeSeek (fromIntegral (l-2))
@@ -320,21 +330,21 @@ createPDFJpeg (JpegFile bits_per_component width height color_space img) = do
         return (PDFReference s) 
     where
        color c = case c of
-           1 -> [(PDFName "ColorSpace",AnyPdfObject $ PDFName "DeviceGray")]
-           3 -> [(PDFName "ColorSpace",AnyPdfObject $ PDFName "DeviceRGB")]
-           4 -> [(PDFName "ColorSpace",AnyPdfObject $ PDFName "DeviceCMYK")
-                ,(PDFName "Decode",AnyPdfObject . map (AnyPdfObject . PDFInteger) $ [1,0,1,0,1,0,1,0])
+           1 -> [entry "ColorSpace" (PDFName "DeviceGray")]
+           3 -> [entry "ColorSpace" (PDFName "DeviceRGB")]
+           4 -> [entry "ColorSpace" (PDFName "DeviceCMYK")
+                ,entry "Decode" (map PDFInteger $ [1,0,1,0,1,0,1,0])
                 ]
            _ -> error "Jpeg color space not supported"
        a' = 
-                 do modifyStrict $ \s -> s  {otherRsrcs = PDFDictionary. M.fromList $ 
-                                                   [ (PDFName "Type",AnyPdfObject . PDFName $ "XObject")
-                                                   , (PDFName "Subtype",AnyPdfObject . PDFName $ "Image")
-                                                   , (PDFName "Width",AnyPdfObject . PDFInteger $ width)
-                                                   , (PDFName "Height",AnyPdfObject . PDFInteger $ height)
-                                                   , (PDFName "BitsPerComponent",AnyPdfObject . PDFInteger $ bits_per_component)
-                                                   , (PDFName "Interpolate", AnyPdfObject True)
-                                                   , (PDFName "Filter",AnyPdfObject . PDFName $ "DCTDecode")
+                 do modifyStrict $ \s -> s  {otherRsrcs = dictFromList $
+                                                   [ entry "Type" (PDFName $ "XObject")
+                                                   , entry "Subtype" (PDFName $ "Image")
+                                                   , entry "Width" (PDFInteger $ width)
+                                                   , entry "Height" (PDFInteger $ height)
+                                                   , entry "BitsPerComponent" (PDFInteger $ bits_per_component)
+                                                   , entry "Interpolate" True
+                                                   , entry "Filter" (PDFName $ "DCTDecode")
                                                    ] ++ color color_space
                                              }
                     tell img        
@@ -352,23 +362,23 @@ createPDFRawImageFromByteString width height interpolate pdfFilter stream =  do
     where     
         getFilter = case pdfFilter of 
                     NoFilter -> []
-                    ASCIIHexDecode -> [(PDFName "Filter",AnyPdfObject . PDFName $ "ASCIIHexDecode")]
-                    ASCII85Decode -> [(PDFName "Filter",AnyPdfObject . PDFName $ "ASCII85Decode")]
-                    LZWDecode -> [(PDFName "Filter",AnyPdfObject . PDFName $ "LZWDecode")]
-                    FlateDecode -> [(PDFName "Filter",AnyPdfObject . PDFName $ "FlateDecode")]
-                    RunLengthDecode -> [(PDFName "Filter",AnyPdfObject . PDFName $ "RunLengthDecode")]
-                    CCITTFaxDecode -> [(PDFName "Filter",AnyPdfObject . PDFName $ "CCITTFaxDecode")]
-                    DCTDecode -> [(PDFName "Filter",AnyPdfObject . PDFName $ "DCTDecode")]
+                    ASCIIHexDecode -> [entry "Filter" (PDFName $ "ASCIIHexDecode")]
+                    ASCII85Decode -> [entry "Filter" (PDFName $ "ASCII85Decode")]
+                    LZWDecode -> [entry "Filter" (PDFName $ "LZWDecode")]
+                    FlateDecode -> [entry "Filter" (PDFName $ "FlateDecode")]
+                    RunLengthDecode -> [entry "Filter" (PDFName $ "RunLengthDecode")]
+                    CCITTFaxDecode -> [entry "Filter" (PDFName $ "CCITTFaxDecode")]
+                    DCTDecode -> [entry "Filter" (PDFName $ "DCTDecode")]
 
         a' =  do 
-                modifyStrict $ \s -> s  {otherRsrcs = PDFDictionary. M.fromList $ 
-                                                   [ (PDFName "Type",AnyPdfObject . PDFName $ "XObject")
-                                                   , (PDFName "Subtype",AnyPdfObject . PDFName $ "Image")
-                                                   , (PDFName "Width",AnyPdfObject . PDFInteger $ width)
-                                                   , (PDFName "Height",AnyPdfObject . PDFInteger $ height)
-                                                   , (PDFName "BitsPerComponent",AnyPdfObject . PDFInteger $ 8)
-                                                   , (PDFName "ColorSpace",AnyPdfObject $ PDFName "DeviceRGB")
-                                                   , (PDFName "Interpolate", AnyPdfObject interpolate)
+                modifyStrict $ \s -> s  {otherRsrcs = dictFromList $
+                                                   [ entry "Type" (PDFName $ "XObject")
+                                                   , entry "Subtype" (PDFName $ "Image")
+                                                   , entry "Width" (PDFInteger $ width)
+                                                   , entry "Height" (PDFInteger $ height)
+                                                   , entry "BitsPerComponent" (PDFInteger $ 8)
+                                                   , entry "ColorSpace" (PDFName "DeviceRGB")
+                                                   , entry "Interpolate" interpolate
                                                    ] ++ getFilter
                                              }
                 tell . fromLazyByteString $ stream
@@ -392,14 +402,14 @@ createPDFRawImageFromARGB width height interpolate stream =  do
         addPixel [] = []
                         
         a' =  do 
-                modifyStrict $ \s -> s  {otherRsrcs = PDFDictionary. M.fromList $ 
-                                                   [ (PDFName "Type",AnyPdfObject . PDFName $ "XObject")
-                                                   , (PDFName "Subtype",AnyPdfObject . PDFName $ "Image")
-                                                   , (PDFName "Width",AnyPdfObject . PDFInteger $  width)
-                                                   , (PDFName "Height",AnyPdfObject . PDFInteger $  height)
-                                                   , (PDFName "BitsPerComponent",AnyPdfObject . PDFInteger $ 8)
-                                                   , (PDFName "ColorSpace",AnyPdfObject $ PDFName "DeviceRGB")
-                                                   , (PDFName "Interpolate", AnyPdfObject interpolate)
+                modifyStrict $ \s -> s  {otherRsrcs = dictFromList $
+                                                   [ entry "Type" (PDFName $ "XObject")
+                                                   , entry "Subtype" (PDFName $ "Image")
+                                                   , entry "Width" (PDFInteger $  width)
+                                                   , entry "Height" (PDFInteger $  height)
+                                                   , entry "BitsPerComponent" (PDFInteger $ 8)
+                                                   , entry "ColorSpace" (PDFName "DeviceRGB")
+                                                   , entry "Interpolate" interpolate
                                                    ]
                                              }
                 tell . fromLazyByteString . B.pack . addPixel . U.toList $ stream  
