@@ -17,17 +17,17 @@
 module Main where
 
 
+import qualified Graphics.PDF.Expression as Expr
 import Graphics.PDF
 import Penrose
 import System.Random
 import qualified Data.Vector.Unboxed as U
 import qualified Data.Text as T
+import Data.Array (listArray)
 import Network.URI 
 import Data.Maybe(fromJust)
 import System.FilePath 
 
-import Paths_HPDF
-
 alpheccarURL = fromJust $ parseURI "http://www.alpheccar.org"
 
 fontDebug :: PDFFont -> T.Text -> Draw ()
@@ -62,12 +62,85 @@ lineStyle  = do
             geometryTest
             
             
-shadingTest :: Draw ()
-shadingTest  = do
-     paintWithShading (RadialShading 0 0 50 0 0 600 (Rgb 1 0 0) (Rgb 0 0 1)) (addShape $ Rectangle 0 (300 :+ 300))
-     paintWithShading (AxialShading 300 300 600 400 (Rgb 1 0 0) (Rgb 0 0 1)) (addShape $ Ellipse 300 300 600 400)
-     
-                     
+shadingTest :: PDFReference PDFPage -> PDF ()
+shadingTest page = do
+    funcObj <-
+        createFunction1Object $
+        Sampled1 $ listArray (0,3) [(1,0,0), (1,0,1), (0,0,1), (0,1,1)]
+    drawWithPage page $ do
+        paintWithShading
+            (RadialShading 0 0 50 0 0 350
+                (ColorFunction1 RGBSpace $ GlobalFunction1 funcObj))
+            (addShape $ Rectangle 0 (300 :+ 300))
+        paintWithShading
+            (AxialShading 300 300 600 400
+                (ColorFunction1 GraySpace $ Interpolated1 1  0.8 0.02))
+            (addShape $ Ellipse 300 300 600 400)
+
+functionalShadingTest :: PDFReference PDFPage -> PDF ()
+functionalShadingTest page = do
+    funcObj <-
+        createFunction2Object $ calculator2 $ \x y ->
+            (1-x,
+             0.5 * (1 + Expr.sinDeg ((360*5) * Expr.sqrt (x*x+y*y))),
+             1-y)
+    drawWithPage page $
+        paintWithShading
+            (FunctionalShading
+                (Matrix 300 0 0 300 150 50)
+                (ColorFunction2 RGBSpace $ GlobalFunction2 funcObj))
+            (addShape $ Rectangle (150:+50) (450 :+ 350))
+
+transparencyTest :: Rectangle -> SoftMask -> Draw ()
+transparencyTest rectB softMask = do
+    let rectA = Rectangle (100:+00) (400 :+ 300)
+    fillColor $ Rgb 1 0 0
+    fill rectA
+
+    paintWithTransparency softMask $ do
+        fillColor $ Rgb 1 1 0
+        fill rectB
+
+    setFillAlpha 0.5
+    let rectC = Rectangle (200:+100) (500 :+ 400)
+    fillColor $ Rgb 0 1 0
+    fill rectC
+
+trafficSign :: AnyFont -> Draw ()
+trafficSign helvetica = do
+    let font = PDFFont helvetica 180
+    let str = "P"
+    strokeColor black
+    drawText $ text font (-0.5 * textWidth font str) (-70) str
+    setWidth 30
+    strokeColor red
+    stroke $ Circle 0 0 100
+    let r = 100 * sqrt 0.5
+    stroke $ Line (-r) (-r) r r
+
+transparencyGroupTest ::
+    AnyFont -> Rectangle -> SoftMask -> PDFReference PDFXForm -> Draw ()
+transparencyGroupTest helvectica rectB softMask trafficSignObj = do
+    applyMatrix $ scale 0.5 0.5
+    applyMatrix $ translate $ 150:+150
+    withNewContext $ trafficSign helvectica
+    withNewContext $ do
+        applyMatrix $ translate $ 300:+0
+        -- alpha channel reveals internal structure of sign - not intended
+        setStrokeAlpha 0.5
+        setFillAlpha 0.5
+        trafficSign helvectica
+    withNewContext $ do
+        applyMatrix $ translate $ 600:+0
+        -- soft mask has no effect on text ... and additional graphics?
+        paintWithTransparency softMask $
+            trafficSign helvectica
+    withNewContext $ do
+        applyMatrix $ translate $ 900:+0
+        paintWithTransparency softMask $
+            drawXObject trafficSignObj
+
+
 patternTest :: PDFReference PDFPage -> PDF ()
 patternTest page = do
      p <- createUncoloredTiling 0 0 100 50 100 50 ConstantSpacing pattern
@@ -326,7 +399,7 @@ instance ParagraphStyle MyVertStyles MyParaStyles  where
                         textStart 0 0
                         setFont (PDFFont theFont fontSize)
                         displayGlyphs (glyph c)
-                    paintWithShading (AxialShading 0 (- getDescent theFont fontSize) w' (getHeight theFont fontSize - getDescent theFont fontSize) (Rgb 1 0 0) (Rgb 0 0 1)) (addShape charRect)
+                    paintWithShading (AxialShading 0 (- getDescent theFont fontSize) w' (getHeight theFont fontSize - getDescent theFont fontSize) (ColorFunction1 RGBSpace $ Interpolated1 1 (1,0,0) (0,0,1))) (addShape charRect)
         in
         (BluePara theFont w', c':l)
     
@@ -656,10 +729,38 @@ testAll timesRoman timesBold helveticaBold symbol zapf jpg = do
         patternTest page7
         
      page8 <- addPage Nothing
-     newSection "Shading" Nothing Nothing $ do
-        drawWithPage page8 $ do
-          shadingTest
-          
+     newSection "Shading" Nothing Nothing $
+        shadingTest page8
+
+     page8a <- addPage Nothing
+     newSection "FunctionalShading" Nothing Nothing $
+        functionalShadingTest page8a
+
+     page8b <- addPage Nothing
+     newSection "Transparency" Nothing Nothing $ do
+        do  let rectB = Rectangle (150:+50) (450 :+ 350)
+            softMask <-
+                createSoftMask rectB
+                    (paintWithShading
+                        (AxialShading 300 300 600 400
+                            (ColorFunction1 GraySpace $
+                             Interpolated1 1  0.8 0.02))
+                        (addShape rectB))
+            drawWithPage page8b $
+                transparencyTest rectB softMask
+
+        page8c <- addPage Nothing
+        do  let bbox = Rectangle (-(150:+150)) (150:+150)
+            softMask <-
+                createSoftMask bbox $ do
+                    fillColor $ Rgb 0.5 0.5 0.5
+                    fill bbox
+            transpGroup <-
+                createTransparencyGroup RGBSpace bbox $
+                trafficSign helveticaBold
+            drawWithPage page8c $
+                transparencyGroupTest helveticaBold bbox softMask transpGroup
+
     page9 <- addPage Nothing
     newSection "Media" Nothing Nothing $ do
       newSection "image" Nothing Nothing $ do   
@@ -688,19 +789,18 @@ testAll timesRoman timesBold helveticaBold symbol zapf jpg = do
     page14 <- addPage Nothing
     rawImage page14
     
-        
 main :: IO()
 main = do
     let rect = PDFRect 0 0 600 400
-    Just timesRoman <- mkStdFont Times_Roman 
-    Just timesBold <- mkStdFont Times_Bold
-    Just helveticaBold <- mkStdFont Helvetica_Bold
-    Just symbol <- mkStdFont Symbol 
-    Just zapf <- mkStdFont ZapfDingbats
+    Right timesRoman <- mkStdFont Times_Roman 
+    Right timesBold <- mkStdFont Times_Bold
+    Right helveticaBold <- mkStdFont Helvetica_Bold
+    Right symbol <- mkStdFont Symbol 
+    Right zapf <- mkStdFont ZapfDingbats
 
-    logoPath <- getDataFileName $ "Test" </> "logo.jpg"
+    let logoPath = "Test/logo.jpg"
     Right jpg <- readJpegFile logoPath
     runPdf "demo.pdf" (standardDocInfo { author= "alpheccar éèçàü", compressed = False}) rect $ do
         testAll timesRoman timesBold helveticaBold symbol zapf jpg
     --print $ charWidth (PDFFont Times_Roman 1) '('
-     
\ No newline at end of file
+     
